import logging
import sys
import json
import time
import os
import importlib.util
import queue
from typing import Optional, Any

import numpy as np
import cv2
import rclpy
from rclpy.node import Node
import std_msgs.msg
import sensor_msgs.msg

from wrapyfi.connect.clients import Client, Clients
from wrapyfi.middlewares.ros2 import ROS2Middleware
from wrapyfi.encoders import JsonEncoder, JsonDecodeHook


class ROS2Client(Client, Node):
    def __init__(self, name: str, in_topic: str, ros2_kwargs: Optional[dict] = None, **kwargs):
        """
        Initialize the client.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param ros2_kwargs: dict: Additional kwargs for the ROS 2 middleware
        :param kwargs: dict: Additional kwargs for the client
        """
        carrier = "tcp"
        if "carrier" in kwargs and kwargs["carrier"] not in ["", None]:
            logging.warning(
                "[ROS 2] ROS 2 currently does not support explicit carrier setting for PUB/SUB pattern. Using TCP.")
        if "carrier" in kwargs:
            del kwargs["carrier"]
        ROS2Middleware.activate(**ros2_kwargs or {})
        Client.__init__(self, name, in_topic, carrier=carrier, **kwargs)
        Node.__init__(self, name + str(hex(id(self))))

    def close(self):
        """
        Close the client.
        """
        if hasattr(self, "_client") and self._client:
            if self._client is not None:
                self.destroy_node()

    def __del__(self):
        self.close()


@Clients.register("NativeObject", "ros2")
class ROS2NativeObjectClient(ROS2Client):
    def __init__(self, name: str, in_topic: str,
                 serializer_kwargs: Optional[dict] = None,
                 deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        The NativeObject listener using the ROS 2 String message assuming the data is serialized as a JSON string.
        Deserializes the data (including plugins) using the decoder and parses it to a Python object.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, in_topic, **kwargs)
        self._client = None
        self._queue = queue.Queue(maxsize=1)

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

    def establish(self):
        """
        Establish the client's connection to the ROS 2 service.
        """
        try:
            from wrapyfi_ros2_interfaces.srv import ROS2NativeObjectService
        except ImportError:
            import wrapyfi
            logging.error("[ROS 2] Could not import ROS2NativeObjectService. "
                          "Make sure the ROS 2 services in wrapyfi_extensions/wrapyfi_ros2_interfaces are compiled. "
                          "Refer to the documentation for more information: \n" +
                          wrapyfi.__doc__ + "ros2_interfaces_lnk.html")
            sys.exit(1)
        self._client = self.create_client(ROS2NativeObjectService, self.in_topic)
        self._req_msg = ROS2NativeObjectService.Request()

        while not self._client.wait_for_service(timeout_sec=1.0):
            logging.info('[ROS 2] Service not available, waiting again...')
        self.established = True

    def request(self, *args, **kwargs):
        """
        Send a request to the ROS 2 service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        :return: Any: The response from the ROS 2 service
        """
        if not self.established:
            self.establish()
        try:
            self._request(*args, **kwargs)
        except Exception as e:
            logging.error("[ROS 2] Service call failed: %s" % e)
        return self._await_reply()

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the ROS 2 service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        """
        # transmit args to server
        args_str = json.dumps([args, kwargs], cls=self._plugin_encoder, **self._plugin_kwargs,
                              serializer_kwrags=self._serializer_kwargs)
        self._req_msg.request = args_str
        future = self._client.call_async(self._req_msg)
        # receive message from server
        while rclpy.ok():
            rclpy.spin_once(self)
            if future.done():
                try:
                    msg = future.result()
                    obj = json.loads(msg.response, object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs)
                    self._queue.put(obj, block=False)
                except Exception as e:
                    logging.error("[ROS 2] Service call failed: %s" % e)
                break

    def _await_reply(self) -> Any:
        """
        Wait for and return the reply from the ROS 2 service.

        :return: Any: The response from the ROS 2 service
        """
        try:
            reply = self._queue.get(block=True)
            return reply
        except queue.Full:
            logging.warning(f"[ROS 2] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__name__}")
            return None


@Clients.register("Image", "ros2")
class ROS2ImageClient(ROS2Client):
    def __init__(self, name: str, in_topic: str, width: int = -1, height: int = -1,
                 rgb: bool = True, fp: bool = False, jpg: bool = False, serializer_kwargs: Optional[dict] = None, **kwargs):
        """
        The Image client using the ROS 2 Image message parsed to a numpy array.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param width: int: The width of the image. Default is -1 (use the width of the received image)
        :param height: int: The height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: Whether the image is RGB. Default is True
        :param fp: bool: Whether to utilize floating-point precision. Default is False
        :param jpg: bool: True if the image should be decompressed from JPG. Default is False
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        super().__init__(name, in_topic, **kwargs)
        self._client = None
        self._queue = queue.Queue(maxsize=1)

        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        if self.fp:
            self._encoding = '32FC3' if self.rgb else '32FC1'
            self._type = np.float32
        else:
            self._encoding = 'bgr8' if self.rgb else 'mono8'
            self._type = np.uint8
        self._pixel_bytes = (3 if self.rgb else 1) * np.dtype(self._type).itemsize

        self._client = None
        self._queue = queue.Queue(maxsize=1)

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

    def establish(self):
        """
        Establish the client's connection to the ROS 2 service.
        """
        try:
            from wrapyfi_ros2_interfaces.srv import ROS2ImageService, ROS2CompressedImageService
        except ImportError:
            import wrapyfi
            logging.error("[ROS 2] Could not import ROS2ImageService. "
                          "Make sure the ROS 2 services in wrapyfi_extensions/wrapyfi_ros2_interfaces are compiled. "
                          "Refer to the documentation for more information: \n" +
                          wrapyfi.__doc__ + "ros2_interfaces_lnk.html")
            sys.exit(1)
        if self.jpg:
            self._client = self.create_client(ROS2CompressedImageService, self.in_topic)
            self._req_msg = ROS2CompressedImageService.Request()
        else:
            self._client = self.create_client(ROS2ImageService, self.in_topic)
            self._req_msg = ROS2ImageService.Request()

        while not self._client.wait_for_service(timeout_sec=1.0):
            logging.info('[ROS 2] Service not available, waiting again...')
        self.established = True

    def request(self, *args, **kwargs):
        """
        Send a request to the ROS 2 service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        :return: Any: The response from the ROS 2 service
        """
        if not self.established:
            self.establish()
        try:
            self._request(*args, **kwargs)
        except Exception as e:
            logging.error("[ROS 2] Service call failed: %s" % e)
        return self._await_reply()

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the ROS 2 service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        """
        # transmit args to server
        args_str = json.dumps([args, kwargs], cls=self._plugin_encoder, **self._plugin_kwargs,
                              serializer_kwrags=self._serializer_kwargs)
        self._req_msg.request = args_str
        future = self._client.call_async(self._req_msg)
        # receive message from server
        while rclpy.ok():
            rclpy.spin_once(self)
            if future.done():
                try:
                    msg = future.result()
                    data = msg.response
                    if self.jpg:
                        self._queue.put((data.format, data.data), block=False)
                    else:
                        self._queue.put((data.height, data.width, data.encoding, data.is_bigendian, data.data),
                                        block=False)
                except Exception as e:
                    logging.error("[ROS 2] Service call failed: %s" % e)
                break

    def _await_reply(self):
        """
        Wait for and return the reply from the ROS 2 service.

        :return: np.array: The received image from the ROS 2 service
        """
        try:
            if self.jpg:
                format, data = self._queue.get(block=True)
                if format != 'jpeg':
                    raise ValueError(f"Unsupported image format: {format}")
                if self.rgb:
                    img = cv2.imdecode(np.frombuffer(data, np.uint8), cv2.IMREAD_COLOR)
                else:
                    img = cv2.imdecode(np.frombuffer(data, np.uint8), cv2.IMREAD_GRAYSCALE)
            else:
                height, width, encoding, is_bigendian, data = self._queue.get(block=True)
                if encoding != self._encoding:
                    raise ValueError("Incorrect encoding for listener")
                if 0 < self.width != width or 0 < self.height != height or len(data) != height * width * self._pixel_bytes:
                    raise ValueError("Incorrect image shape for listener")
                img = np.frombuffer(data, dtype=np.dtype(self._type).newbyteorder('>' if is_bigendian else '<')).reshape((height, width, -1))
                if img.shape[2] == 1:
                    img = img.squeeze(axis=2)
            return img
        except queue.Full:
            logging.warning(f"[ROS 2] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__name__}")
            return None


@Clients.register("AudioChunk", "ros2")
class ROS2AudioChunkClient(ROS2Client):
    def __init__(self, name: str, in_topic: str,
                 channels: int = 1, rate: int = 44100, chunk: int = -1,
                 serializer_kwargs: Optional[dict] = None, **kwargs):
        """
        The AudioChunk client using the ROS 2 Audio message parsed to a numpy array.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param channels: int: Number of channels in the audio. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: Number of samples in the audio chunk. Default is -1 (use the chunk size of the received audio)
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        super().__init__(name, in_topic, **kwargs)
        self._client = None
        self._queue = queue.Queue(maxsize=1)

        self.channels = channels
        self.rate = rate
        self.chunk = chunk

        self._client = None
        self._queue = queue.Queue(maxsize=1)

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

    def establish(self):
        """
        Establish the client's connection to the ROS 2 service.
        """
        try:
            from wrapyfi_ros2_interfaces.srv import ROS2AudioService
        except ImportError:
            import wrapyfi
            logging.error("[ROS 2] Could not import ROS2AudioService. "
                          "Make sure the ROS 2 services in wrapyfi_extensions/wrapyfi_ros2_interfaces are compiled. "
                          "Refer to the documentation for more information: \n" +
                          wrapyfi.__doc__ + "ros2_interfaces_lnk.html")
            sys.exit(1)
        self._client = self.create_client(ROS2AudioService, self.in_topic)
        self._req_msg = ROS2AudioService.Request()

        while not self._client.wait_for_service(timeout_sec=1.0):
            logging.info('[ROS 2] Service not available, waiting again...')
        self.established = True

    def request(self, *args, **kwargs):
        """
        Send a request to the ROS 2 service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        :return: Any: The response from the ROS 2 service
        """
        if not self.established:
            self.establish()
        try:
            self._request(*args, **kwargs)
        except Exception as e:
            logging.error("[ROS 2] Service call failed: %s" % e)
        return self._await_reply()

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the ROS 2 service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        """
        args_str = json.dumps([args, kwargs], cls=self._plugin_encoder, **self._plugin_kwargs,
                              serializer_kwrags=self._serializer_kwargs)
        self._req_msg.request = args_str
        future = self._client.call_async(self._req_msg)

        while rclpy.ok():
            rclpy.spin_once(self)
            if future.done():
                try:
                    msg = future.result()
                    data = msg.response
                    self._queue.put((data.chunk_size, data.channels, data.sample_rate, data.encoding, data.is_bigendian, data.data),
                                        block=False)
                except Exception as e:
                    logging.error("[ROS 2] Service call failed: %s" % e)
                break

    def _await_reply(self):
        """
        Wait for and return the reply from the ROS 2 service.

        :return: Tuple[np.ndarray, int]: The received message as a numpy array formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        try:
            chunk, channels, rate, encoding, is_bigendian, data = self._queue.get(block=False)
            if 0 < self.rate != rate:
                raise ValueError("Incorrect audio rate for publisher")
            if encoding not in ['S16LE', 'S16BE']:
                raise ValueError("Incorrect encoding for listener")
            if 0 < self.chunk != chunk or self.channels != channels or len(data) != chunk * channels * 4:
                raise ValueError("Incorrect audio shape for listener")
            aud = np.frombuffer(data, dtype=np.dtype(np.float32).newbyteorder('>' if is_bigendian else '<')).reshape(
                (chunk, channels))
            # aud = aud / 32767.0
            return aud, rate
        except queue.Full:
            logging.warning(f"[ROS 2] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__name__}")
            return None




"""
Encoder and Decoder for TensorFlow Tensor Data via Wrapyfi.

This script provides mechanisms to encode and decode TensorFlow tensor data using Wrapyfi.
It utilizes base64 encoding to convert binary data into ASCII strings.

The script contains a class, `TensorflowTensor`, registered as a plugin to manage the
conversion of TensorFlow tensor data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - TensorFlow: An end-to-end open-source platform for machine learning (refer to https://www.tensorflow.org/install for installation instructions)
        Note: If TensorFlow is not available, HAVE_TENSORFLOW will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install tensorflow``  # Basic installation of TensorFlow
"""

import io
import base64

import numpy as np

from wrapyfi.utils import *

try:
    import tensorflow
    HAVE_TENSORFLOW = True
except ImportError:
    HAVE_TENSORFLOW = False


@PluginRegistrar.register(types=None if not HAVE_TENSORFLOW else tensorflow.Tensor.__mro__[:-1])
class TensorflowTensor(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the TensorflowTensor plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode TensorFlow tensor data into a base64 ASCII string.

        :param obj: tensorflow.Tensor: The TensorFlow tensor data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name and encoded data string
        """
        with io.BytesIO() as memfile:
            np.save(memfile, obj.numpy())
            obj_data = base64.b64encode(memfile.getvalue()).decode('ascii')
        return True, dict(__wrapyfi__=(str(self.__class__.__name__), obj_data))

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into TensorFlow tensor data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, tensorflow.Tensor]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - tensorflow.Tensor: The decoded TensorFlow tensor data
        """
        with io.BytesIO(base64.b64decode(obj_full[1].encode('ascii'))) as memfile:
            return True, tensorflow.convert_to_tensor(np.load(memfile))



"""
Channeling Example using Wrapyfi.

This script demonstrates message channeling through three different middleware (A, B, and C) using
the MiddlewareCommunicator within the Wrapyfi library. It allows message publishing and listening
functionalities between processes or machines.

Demonstrations:
    - Using NativeObject, Image, and AudioChunk messages
    - Transmitting messages through different middlewares (e.g., ROS, YARP, ZeroMQ)
    - Applying the PUB/SUB pattern with channeling

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - NumPy: Used for creating arrays (installed with Wrapyfi)

    Install using pip:
        ``pip install numpy``

Run:
    # On machine 1 (or process 1): Mode: Publisher sends messages through channels A [ZeroMQ], B [ROS 2], and C [YARP]
        ``python3 channeling_example.py --mode publish --mware_A zeromq --mware_B ros2 --mware_C yarp``

    # On machine 2 (or process 2): Mode: Listener waits for message from channels A [ZeroMQ], B [ROS 2], and C [YARP]
        ``python3 channeling_example.py --mode listen --mware_A zeromq --mware_B ros2 --mware_C yarp``

    # On machine 3 (or process 3): Mode: Listener waits for message from channel C [YARP]
        ``python3 channeling_example.py --mode listen --mware_C yarp``

    # On machine 4 (or process 4) [OPTIONAL]: Mode: Listener waits for message from channel B [ROS 2]
        ``python3 channeling_example.py --mode listen --mware_B ros2``
"""

import argparse
import time

import numpy as np

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class ChannelingCls(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject", "$mware_A", "ChannelingCls", "/example/native_A_msg",
        carrier="mcast", should_wait=True
    )
    @MiddlewareCommunicator.register(
        "Image", "$mware_B", "ChannelingCls", "/example/image_B_msg",
        carrier="tcp", width="$img_width", height="$img_height", rgb=True, should_wait=False
    )
    @MiddlewareCommunicator.register(
        "AudioChunk", "$mware_C", "ChannelingCls", "/example/audio_C_msg",
        carrier="tcp", rate="$aud_rate", chunk="$aud_chunk", channels="$aud_channels", should_wait=False
    )
    def read_mulret_mulmware(self, img_width=200, img_height=200,
                             aud_rate=44100, aud_chunk=8820, aud_channels=1,
                             mware_A=None, mware_B=None, mware_C=None):
        """Read and forward messages through channels A, B, and C."""
        ros_img = np.random.randint(256, size=(img_height, img_width, 3), dtype=np.uint8)
        zeromq_aud = (np.random.uniform(-1, 1, aud_chunk), aud_rate,)
        yarp_native = [ros_img, zeromq_aud]
        return yarp_native, ros_img, zeromq_aud


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Channeling Example using Wrapyfi.")
    parser.add_argument(
        "--mode", type=str, default="publish",
        choices={"publish", "listen"},
        help="The transmission mode"
    )
    parser.add_argument(
        "--mware_A", type=str, default="none",
        choices=MiddlewareCommunicator.get_communicators().update({"none"}),
        help="The middleware to use for transmission of channel A"
    )
    parser.add_argument(
        "--mware_B", type=str, default="none",
        choices=MiddlewareCommunicator.get_communicators().update({"none"}),
        help="The middleware to use for transmission of channel B"
    )
    parser.add_argument(
        "--mware_C", type=str, default="none",
        choices=MiddlewareCommunicator.get_communicators().update({"none"}),
        help="The middleware to use for transmission of channel C"
    )
    return parser.parse_args()


def main(args):
    """Main function to initiate ChannelingCls class and communication."""
    channeling = ChannelingCls()
    channeling.activate_communication(channeling.read_mulret_mulmware, mode=args.mode)

    while True:
        A_native, B_img, C_aud = channeling.read_mulret_mulmware(
            mware_A=args.mware_A, mware_B=args.mware_B, mware_C=args.mware_C
        )
        if A_native is not None:
            print(f"{args.mware_A}_native::", A_native)
        if B_img is not None:
            print(f"{args.mware_B}_img::", B_img)
        if C_aud is not None:
            print(f"{args.mware_C}_aud::", C_aud)
        time.sleep(0.1)


if __name__ == "__main__":
    args = parse_args()
    main(args)



import argparse

import cv2

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR
from wrapyfi.config.manager import ConfigManager


class ExperimentController(MiddlewareCommunicator):
    WEBCAM_HEIGHT = 240
    WEBCAM_WIDTH = 320
    ICUB_CAM_HEIGHT = 240
    ICUB_CAM_WIDTH = 320
    SIXDREPNET_CAM_HEIGHT = 240
    SIXDREPNET_CAM_WIDTH = 320

    def __init__(self, **kwargs):
        super(ExperimentController, self).__init__()

    @MiddlewareCommunicator.register("NativeObject", "$_sixdrepnet_mware", "ExperimentController",
                                     "/control_interface/orientation_sixdrepnet", should_wait=False)
    @MiddlewareCommunicator.register("NativeObject", "$_waveshareimu_mware", "ExperimentController",
                                     "/control_interface/orientation_waveshareimu", should_wait=False)
    def listen_orientation_sixdrepnet_waveshareimu(self, _sixdrepnet_mware="yarp", _waveshareimu_mware="yarp"):
        return None, None,

    @MiddlewareCommunicator.register("NativeObject", "$_mware", "ExperimentController",
                                     "/control_interface/orientation_icub", should_wait=False)
    def publish_orientation_icub(self, obj, _mware="yarp"):
        return obj,

    @MiddlewareCommunicator.register("NativeObject", "$_mware", "ExperimentController",
                                     "/control_interface/gaze_pupil", should_wait=False)
    def listen_gaze_pupil(self, _mware="zeromq"):
        return None,

    @MiddlewareCommunicator.register("NativeObject", "$_mware", "ExperimentController",
                                     "/control_interface/gaze_icub", should_wait=False)
    def publish_gaze_icub(self, obj, _mware="yarp"):
        return obj,


    @MiddlewareCommunicator.register("Image", "$_mware", "ExperimentController",
                                     "/control_interface/image_webcam",
                                     width="$_width", height="$_height", rgb=True, fp=False, should_wait=False,
                                     jpg=True)
    def listen_image_webcam(self, _mware="ros2", _width=WEBCAM_WIDTH, _height=WEBCAM_HEIGHT):
        return None,

    @MiddlewareCommunicator.register("Image", "$_mware", "ExperimentController",
                                     "/control_interface/image_sixdrepnet",
                                     width="$_width", height="$_height", rgb=True, fp=False, should_wait=False)
    def publish_sixdrepnet_cam(self, img, _mware="ros2", _width=SIXDREPNET_CAM_WIDTH, _height=SIXDREPNET_CAM_HEIGHT):
        # modify size of image
        if img is not None:
            img = cv2.resize(img, (self.SIXDREPNET_CAM_WIDTH, self.SIXDREPNET_CAM_HEIGHT))
        return img,

    def priority_control_sources(self, orientation_sixdrepnet, orientation_imu, control_sources):
        if control_sources[0] == "vision":
            if orientation_sixdrepnet is not None:
                # print("Orientation 6DRepNet: ", orientation_sixdrepnet)
                self.publish_orientation_icub(orientation_sixdrepnet)
            elif orientation_imu is not None and "imu" in control_sources:
                # print("Orientation IMU: ", orientation_imu)
                self.publish_orientation_icub(orientation_imu)
        elif control_sources[0] == "imu":
            if orientation_imu is not None:
                # print("Orientation IMU: ", orientation_imu)
                self.publish_orientation_icub(orientation_imu)
            elif orientation_sixdrepnet is not None and "vision" in control_sources:
                # print("Orientation 6DRepNet: ", orientation_sixdrepnet)
                self.publish_orientation_icub(orientation_sixdrepnet)

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--wrapyfi_cfg",
                        help="File to load Wrapyfi configs for running instance. "
                             "Choose one of the configs available in "
                             "./wrapyfi_configs/gaze_mirroring_multisensor. "
                             "for each running instance. All configs with a prefix of COMP must "
                             "have corresponding instances running. OPT prefixed config is optional "
                             "(only when using a vision model) executes the script on a machine connected "
                             "to the camera (or has access to the camera topic).",
                        type=str)
    parser.add_argument("--control_sources",
                        help="Control sources to use for the experiment. The order of sources indicates the priority. "
                             "For example, if vision is the first source, then the vision source will "
                             "be used for control. If vision is not available, then the IMU source will be used. "
                             "If one source is provided, then the experiment will run with that source. ",
                        type=str, default=["vision", "imu"], nargs="+", choices=["vision", "imu"])
    parser.add_argument("--enable_gaze",
                        help="Enable the gaze (eye movement) control. If not enabled, "
                             "then the gaze control will not be used. ",
                        action="store_true", default=False)

    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    if args.wrapyfi_cfg:
        print("Wrapyfi config loading from ", args.wrapyfi_cfg)
        ConfigManager(args.wrapyfi_cfg)

    ec = ExperimentController()
    image_cam = None
    while True:
        if "vision" in args.control_sources:
            image_webcam, = ec.listen_image_webcam()
            image_webcam_linked, = ec.publish_sixdrepnet_cam(image_cam)
            if image_webcam is None:
                image_webcam = image_webcam_linked
            image_cam = image_webcam
            if image_webcam is not None:
                cv2.imshow("Webcam image", image_webcam)
                cv2.waitKey(1)

        orientation_sixdrepnet, orientation_imu = ec.listen_orientation_sixdrepnet_waveshareimu()
        ec.priority_control_sources(orientation_sixdrepnet, orientation_imu, args.control_sources)

        if args.enable_gaze:
            gaze_pupil, = ec.listen_gaze_pupil()
            if gaze_pupil is not None:
                # print("Gaze Pupil: ", gaze_pupil)
                ec.publish_gaze_icub(gaze_pupil)







"""
A simple example of using the Wrapper class to publish and listen to audio and video streams.

This script demonstrates the capability to transmit audio and video streams using the 
MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern
allowing message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the Image and AudioChunk messages
    - Transmitting an audio and video stream
    - Applying the PUB/SUB pattern with persistence
    - Transmitting OpenCV images captured from a camera on the publishing end and displayed on the listener's end
    - Transmitting a sounddevice (PortAudio with NumPy) audio chunk captured from a microphone which can be played back on the listener's end
    - Spawning of multiple processes specifying different functionality for listeners and publishers
    - Using a single return wrapper functionality in conjunction with synchronous callbacks

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - OpenCV: Used for handling and creating images (installed with Wrapyfi)
    - sounddevice, soundfile: Used for handling audio

    Install using pip:
        ``pip install sounddevice soundfile``

Run:
    # Alternative 1: Separate audio and video publishing
        # On machine 1 (or process 1): The audio stream publishing
        
        ``python3 cam_mic.py --mode publish --stream audio --mic_source 0``
        
        # On machine 2 (or process 2): The video stream publishing
        
        ``python3 cam_mic.py --mode publish --stream video --img_source 0``
        
        # On machine 3 (or process 3): The audio stream listening
        
        ``python3 cam_mic.py --mode listen --stream audio``
        
        # On machine 4 (or process 4): The video stream listening
        
        ``python3 cam_mic.py --mode listen --stream video``
        
    # Alternative 2: Concurrent audio and video publishing
        # On machine 1 (or process 1): The audio/video stream publishing
        
        ``python3 cam_mic.py --mode publish --stream audio video --img_source 0 --mic_source 0``
        
        # On machine 2 (or process 2): The audio/video stream listening
        
        ``python3 cam_mic.py --mode listen --stream audio video``
        
"""

import logging
import argparse

import cv2
import sounddevice as sd
import numpy as np

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class CamMic(MiddlewareCommunicator):
    __registry__ = {}

    def __init__(self, *args, stream=("audio", "video"), mic_source=0,
                 mic_rate=44100, mic_chunk=10000, mic_channels=1, img_source=0,
                 img_width=320, img_height=240, mware=None, **kwargs):
        super(MiddlewareCommunicator, self).__init__()
        self.mic_source = mic_source
        self.mic_rate = mic_rate
        self.mic_chunk = mic_chunk
        self.mic_channels = mic_channels

        self.img_source = img_source
        self.img_width = img_width
        self.img_height = img_height

        self.enable_audio = "audio" in stream
        self.enable_video = self.vid_cap = "video" in stream

        self.mware = mware

    @MiddlewareCommunicator.register("Image", "$mware", "CamMic", "/cam_mic/cam_feed",
                                     carrier="", width="$img_width", height="$img_height", rgb=True, jpg=True,
                                     queue_size=10)
    def collect_cam(self, img_width=320, img_height=240, mware=None):
        """Collect images from the camera."""
        if self.vid_cap is True:
            self.vid_cap = cv2.VideoCapture(self.img_source)
            if img_width > 0 and img_height > 0:
                self.vid_cap.set(cv2.CAP_PROP_FRAME_WIDTH, img_width)
                self.vid_cap.set(cv2.CAP_PROP_FRAME_HEIGHT, img_height)
            if not self.vid_cap.isOpened():
                self.vid_cap.release()
                self.vid_cap = False
        if self.vid_cap:
            grabbed, img = self.vid_cap.read()
            if grabbed:
                print("Video frame grabbed")
            else:
                print("Video frame not grabbed")
                img = np.random.randint(256, size=(img_height, img_width, 3), dtype=np.uint8)
        else:
            print("Video capturer not opened")
            img = np.random.randint(256, size=(img_height, img_width, 3), dtype=np.uint8)
        return img,

    @MiddlewareCommunicator.register("AudioChunk", "$mware", "CamMic", "/cam_mic/audio_feed",
                                     carrier="", rate="$mic_rate", chunk="$mic_chunk", channels="$mic_channels")
    def collect_mic(self, aud=None, mic_rate=44100, mic_chunk=int(44100 / 5), mic_channels=1, mware=None):
        """Collect audio from the microphone."""
        aud = aud, mic_rate
        return aud,

    def capture_cam_mic(self):
        """Capture audio and video from the camera and microphone."""
        if self.enable_audio:
            # capture the audio stream from the microphone
            with sd.InputStream(device=self.mic_source, channels=self.mic_channels, callback=self._mic_callback,
                                blocksize=self.mic_chunk, samplerate=self.mic_rate):
                while True:
                    pass
        elif self.enable_video:
            while True:
                self.collect_cam(mware=self.mware)

    def _mic_callback(self, audio, frames, time, status):
        """Callback for the microphone audio stream."""
        if self.enable_video:
            self.collect_cam(img_width=self.img_width, img_height=self.img_height, mware=self.mware)
        self.collect_mic(audio, mic_rate=self.mic_rate, mic_chunk=self.mic_chunk, mic_channels=self.mic_channels, mware=self.mware)
        print(audio.flatten(), audio.min(), audio.mean(), audio.max())

    def __del__(self):
        """Release the video capture device."""
        if not isinstance(self.vid_cap, bool):
            self.vid_cap.release()


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="A streamer and listener for audio and video streams.")
    parser.add_argument("--mode", type=str, default="publish", choices={"publish", "listen"}, help="The transmission mode")
    parser.add_argument("--mware", type=str, default=DEFAULT_COMMUNICATOR, choices=MiddlewareCommunicator.get_communicators(),
                                                                             help="The middleware to use for transmission")
    parser.add_argument("--stream", nargs="+", default=["video", "audio"],
                        choices={"video", "audio"},
                        help="The streamed sensor data")
    parser.add_argument("--img_source", type=int, default=0, help="The video capture device id (int camera id)")
    parser.add_argument("--img_width", type=int, default=320, help="The image width")
    parser.add_argument("--img_height", type=int, default=240, help="The image height")
    parser.add_argument("--mic_source", type=int, default=None, help="The audio capture device id (int microphone id from python3 -m sounddevice)")
    parser.add_argument("--mic_rate", type=int, default=44100, help="The audio sampling rate")
    parser.add_argument("--mic_channels", type=int, default=1, help="The audio channels")
    parser.add_argument("--mic_chunk", type=int, default=10000, help="The transmitted audio chunk size")
    parser.add_argument("--sound_device", type=int, default=0, help="The sound device to use for audio playback")
    parser.add_argument("--list_sound_devices", action="store_true", help="List all available sound devices and exit")
    return parser.parse_args()


def sound_play(my_aud, blocking=True, device=0):
    """
    Play audio using sounddevice.

    :param my_aud: Tuple[np.ndarray, int]: The audio chunk and sampling rate to play
    :param blocking: bool: Whether to block the execution until the audio is played
    :param device: int: The sound device to use for audio playback
    :return: bool: Whether the audio was played successfully
    """
    try:
        sd.play(*my_aud, blocking=blocking, device=device)
        return True
    except sd.PortAudioError:
        logging.warning("PortAudioError: No device is found or the device is already in use. Will try again in 3 seconds.")
        return False


def main(args):
    if args.list_sound_devices:
        print(sd.query_devices())
        return

    cam_mic = CamMic(stream=args.stream, mic_source=args.mic_source, mware=args.mware)

    if args.mode == "publish":
        cam_mic.activate_communication(CamMic.collect_cam, mode="publish")
        cam_mic.activate_communication(CamMic.collect_mic, mode="publish")
        cam_mic.capture_cam_mic()
    elif args.mode == "listen":
        cam_mic.activate_communication(CamMic.collect_cam, mode="listen")
        cam_mic.activate_communication(CamMic.collect_mic, mode="listen")
        while True:
            if "audio" in args.stream:
                (aud, mic_rate), = cam_mic.collect_mic(mic_rate=args.mic_rate, mic_chunk=args.mic_chunk, mic_channels=args.mic_channels, mware=args.mware)
            else:
                aud = mic_rate = None
            if "video" in args.stream:
                img, = cam_mic.collect_cam(img_source=args.img_source, img_width=args.img_width, img_height=args.img_height, mware=args.mware)
            else:
                img = None
            if img is not None:
                cv2.imshow("Received image", img)
                cv2.waitKey(1)
            if aud is not None:
                sound_play((aud, mic_rate), blocking=False, device=args.sound_device)
                sd.wait(1)


if __name__ == "__main__":
    args = parse_args()
    main(args)



import logging
import os
from glob import glob

from wrapyfi.utils import dynamic_module_import


class Clients(object):
    """
    A class that holds all clients and their corresponding middleware communicators.
    """
    registry = {}
    mwares = set()

    @classmethod
    def register(cls, data_type: str, communicator: str):
        """
        Register a client with the given data type and middleware communicator.

        :param data_type: str: The data type to register the client for e.g., "NativeObject", "Image", "AudioChunk", etc.
        :param communicator: str: The middleware communicator to register the client for e.g., "ros", "ros2", "yarp", "zeromq", etc.
        """
        def decorator(cls_):
            cls.registry[data_type + ":" + communicator] = cls_
            cls.mwares.add(communicator)
            return cls_
        return decorator

    @staticmethod
    def scan():
        """
        Scan for clients and add them to the registry.
        """
        modules = glob(os.path.join(os.path.dirname(__file__), "..", "clients", "*.py"), recursive=True)
        modules = ["wrapyfi.clients." + module.replace(os.path.dirname(__file__) + "/../clients/", "") for module in
                   modules]
        dynamic_module_import(modules, globals())


class Client(object):
    """
    A base class for clients.
    """
    def __init__(self, name: str, in_topic: str, carrier: str = "", **kwargs):
        """
        Initialize the client.

        :param name: str: The name of the client
        :param in_topic: str: The topic to listen to
        :param carrier: str: The middleware carrier to use
        """
        self.__name__ = name
        self.in_topic = in_topic
        self.carrier = carrier
        self.established = False

    def establish(self):
        """
        Establish the client.
        """
        raise NotImplementedError

    def request(self, *args, **kwargs):
        """
        Send a request to the server.
        """
        raise NotImplementedError

    def _request(self, *args, **kwargs):
        """
        Internal method for sending a request to the server in the background.
        """
        raise NotImplementedError

    def _await_reply(self):
        """
        Internal method for awaiting a reply from the server in the background.
        """
        raise NotImplementedError

    def close(self):
        """
        Close the connection.
        """
        raise NotImplementedError
